"""
Клиент для http-запросов к сервисам expressmoney c автоматической авторизацией в Identity-Aware Proxy
"""

__all__ = ('Request',)

import os
import abc
import datetime
import re
from contextlib import suppress
from typing import Union
from urllib.parse import urlencode

import requests
from google.auth.transport.requests import Request as GoogleRequest
from google.oauth2 import id_token


def is_informational(code):
    return 100 <= code <= 199


def is_success(code):
    return 200 <= code <= 299


def is_redirect(code):
    return 300 <= code <= 399


def is_client_error(code):
    return 400 <= code <= 499


def is_server_error(code):
    return 500 <= code <= 599


class _HttpClient(abc.ABC):
    """Abstract HTTP client"""

    _client = None
    _project = 'expressmoney' if os.getenv('IS_PROD_REQUESTS') or os.getenv('GAE_APPLICATION') else 'expressmoney-dev-1'

    def __init__(self,
                 service: str = None,
                 path: str = '/',
                 query_params: Union[None, dict] = None,
                 access_token: str = None,
                 timeout: tuple = (30, 30),
                 version: str = None,
                 ):
        """
        Common params for all http clients
        Args:
            service: 'default'
            path: '/user'
            access_token: 'Bearer DFD4345345D'
        """
        self._service = service
        self._path = path
        self._query_params = query_params
        self._access_token = access_token
        self._timeout = timeout
        self._version = version

    @abc.abstractmethod
    def get(self):
        pass

    @abc.abstractmethod
    def post(self, payload: dict):
        pass

    @abc.abstractmethod
    def put(self, payload: dict):
        pass


class Request(_HttpClient):
    """Base HTTP Client"""

    def get(self, url=None):
        response = requests.get(url if url else self.url, headers=self._headers, timeout=self._timeout)
        return response

    def delete(self):
        response = requests.delete(self.url, headers=self._headers, timeout=self._timeout)
        return response

    def post(self, payload: dict):
        response = requests.post(self.url, json=payload, headers=self._headers, timeout=self._timeout)
        return response

    def put(self, payload: dict = None):
        payload = {} if payload is None else payload
        response = requests.put(self.url, json=payload, headers=self._headers, timeout=self._timeout)
        return response

    def post_file(self, file, file_name: str, type_: int = 1, is_public=False):
        """
        Save file in Google Storage
        Args:
            file: BytesIO file
            file_name: "name_file.pdf"
            type_: 1 - other files. All types see in storage service
            is_public: True - access to file without auth.

        Returns:

        """
        if len(file_name.split('.')) == 0:
            raise Exception('File name in format "name_file.pdf"')

        ext = file_name.split('.')[-1]
        name = ''
        for value in file_name.split('.')[0:-1]:
            name += value
        name = f'{name}_{datetime.datetime.now().timestamp()}'
        name = re.sub('[^0-9a-zA-Z_]', '', name)
        new_file_name = f'{name}.{ext}'
        data = {
            'name': name,
            'type': type_,
            'is_public': is_public,

        }

        with suppress(Exception):
            file = getattr(file, 'file')

        response = requests.post(
            url=self.url,
            data=data,
            files={"file": (new_file_name, file)},
            headers=self._headers,
            timeout=self._timeout
        )
        if not any((is_success(response.status_code), is_client_error(response.status_code))):
            try:
                raise Exception(f'{response.status_code}:{response.url}:{response.json()}')
            except Exception:
                raise Exception(f'{response.status_code}:{response.url}:{response.text}')
        return response

    @property
    def url(self):
        local_url = 'http://127.0.0.1:8000'
        domain = f'https://{self._service}-dot-{self._project}.appspot.com' if self._service else local_url
        url = f'{domain}{self._path}'
        url_with_params = url if self._query_params is None else f'{url}?{urlencode(self._query_params)}'
        return url_with_params

    @property
    def _headers(self):
        headers = dict()
        headers.update(self._get_authorization())
        if self._version:
            headers.update({'Accept': f'application/json; version={self._version}'})
        return headers

    def _get_authorization(self) -> dict:
        authorization = {'X-Forwarded-Authorization': f'Bearer {self._access_token}'} if self._access_token else {}
        open_id_connect_token = id_token.fetch_id_token(GoogleRequest(), self._get_iap_client_id())
        iap_token = {'Authorization': f'Bearer {open_id_connect_token}'}
        authorization.update(iap_token)
        return authorization

    @staticmethod
    def _get_iap_client_id():
        return '1086735462412-c11726ttrnh7t33elp4gciine1uog21a.apps.googleusercontent.com'



"""
Клиенты для простого доступа к Google Cloud API из приложений Django.
Альтернатива  тяжелым пакетам от Google.
"""

import base64
import json
import os

import requests
from django.utils.module_loading import import_string
from google.auth import default
from google.auth.transport.requests import AuthorizedSession
from rest_framework import status
from rest_framework.status import is_success, HTTP_404_NOT_FOUND


class GoogleCloudServiceError(Exception):
    pass


class VertexAIError(GoogleCloudServiceError):
    pass


def download_object(object_name: str, return_url: bool = False):
    """
    Скачивает файлы с Google Cloud Storage
    @param object_name: Пример pets/dog.png, Символ / заменить на %2F pets%2Fdog.png.
                        Символы требующие замены смотреть здесь
                        https://cloud.google.com/storage/docs/request-endpoints#encoding
    @return: Бинарный файл
    """
    bucket_name = 'expressmoney'
    if object_name[:17] == 'gs://expressmoney':
        object_name = object_name[18:]
    object_name = object_name.replace('/', '%2F')
    url = f'https://storage.googleapis.com/storage/v1/b/{bucket_name}/o/{object_name}?alt=media'
    credentials, project = default(scopes=['https://www.googleapis.com/auth/cloud-platform'])
    authed_session = AuthorizedSession(credentials)
    response = authed_session.get(url)
    if not is_success(response.status_code):
        if response.status_code == HTTP_404_NOT_FOUND:
            raise GoogleCloudServiceError(response.text)
        else:
            raise GoogleCloudServiceError(response.json())
    if return_url:
        return response.content, url
    else:
        return response.content


def get_secret(secret_key):
    endpoint = 'https://secretmanager.googleapis.com'
    project = "1086735462412" if not os.getenv('PROJECT') else os.getenv('PROJECT')
    uri = f'/v1/projects/{project}/secrets/{secret_key}/versions/latest:access'
    url = f'{endpoint}{uri}'
    credentials, project = default(scopes=['https://www.googleapis.com/auth/cloud-platform'])
    authed_session = AuthorizedSession(credentials)
    response = authed_session.get(url)
    if not is_success(response.status_code):
        raise GoogleCloudServiceError(response.json())
    secret_bytes = response.json().get('payload').get('data')
    secret = base64.b64decode(secret_bytes).decode('utf-8')
    return secret


class VertexAI:
    """Google Cloud VertexAI """

    REGION = 'europe-west1'
    PROJECT_ID = '1086735462412'

    def get_score(self):
        if self._action == 'explain':
            self._set_explanations()
        response = self._request()
        predictions = response.json().get('predictions')[0]
        for target_index, target_value in enumerate(predictions['classes']):
            if target_value == 'true':
                score = predictions.get('scores')[target_index]
                return score

        raise VertexAIError("Not found target 'true' in prediction result")

    def _request(self):
        if self.__response is None:
            data = dict()
            data['instances'] = []
            data['instances'].append(self._features_sample)
            headers = {
                'Content-Type': 'application/json',
                'Authorization': f'Bearer {self.bearer}',
            }
            self.__response = requests.post(self.url, json=data, headers=headers)
            if not status.is_success(self.__response.status_code):
                error = f'VertexAI prediction error: {self.__response.status_code}: {self.__response.text}'
                raise VertexAIError(error)
        return self.__response

    def _set_explanations(self):
        """Получить значимость фичей в расчете скора"""
        response = self._request().json()
        target_name = response.get('explanations')[0].get('attributions')[0].get('outputDisplayName')
        explanations = response.get('explanations')[0].get('attributions')[0].get('featureAttributions')
        multiplier = 1000 if target_name == 'true' else -1000
        explanations = {key: multiplier * value for key, value in explanations.items()}
        explanations = {k: v for k, v in sorted(explanations.items(), key=lambda item: item[1], reverse=True)}
        self.explanations = explanations

    @staticmethod
    def _get_token():
        if os.getenv('GAE_APPLICATION'):
            # when we are in GAE env:
            metadata_url = 'http://metadata.google.internal/computeMetadata/v1/'
            metadata_headers = {'Metadata-Flavor': 'Google'}
            service_account = 'default'

            url = '{}instance/service-accounts/{}/token'.format(
                metadata_url, service_account)

            # Request an access token from the metadata server.
            r = requests.get(url, headers=metadata_headers)
            r.raise_for_status()

            # Extract the access token from the response.
            access_token = r.json()['access_token']
        else:
            # when running locally:
            stream = os.popen('gcloud auth application-default print-access-token')
            access_token = stream.read().strip()

        return access_token

    def __init__(self, endpoint_id, features_sample: dict, action='predict'):
        self.explanations = None
        self._action = action
        self.endpoint_id = endpoint_id
        self.url = f'https://{self.REGION}-aiplatform.googleapis.com/v1/projects/{self.PROJECT_ID}/locations/' \
                   f'{self.REGION}/endpoints/{self.endpoint_id}:{action}'
        self.bearer = self._get_token()
        self._features_sample = features_sample
        self.__response = None


class Tasks:
    """Google Cloud Tasks"""

    SERVICE_ENDPOINT = 'https://cloudtasks.googleapis.com'
    SERVICE_ENDPOINT_URI = '/v2/projects/expressmoney/locations/europe-west1/queues/default/tasks/'

    def __init__(self, service: str, uri: str, user: None, ):
        """
        Укажите параметры точки сервиса Expressmoney к которому будет направлен запрос.
        Args:
            service (str): Название сервис ExpressMoney. Пример: loans
            service (str): Путь до точки. Пример: /orders/order . Слэш в начале и нет в конце
            user (None, int, User): От имени какого юзера будет авторизован запрос.
        """

        credentials, _ = default(scopes=['https://www.googleapis.com/auth/cloud-platform'])
        self.authed_session = AuthorizedSession(credentials)
        self._response = None
        self._create_task(service, uri)
        self._add_access_token(user)

    def post(self, payload: dict):
        self._add_http_method('POST')
        self._add_body(payload)
        self._request()

    def get(self):
        self._add_http_method('GET')
        self._request()

    def put(self, payload: dict):
        self._add_http_method('PUT')
        self._add_body(payload)
        self._request()

    def _request(self):
        self._response = self.authed_session.post(self._servie_url, json.dumps(self._task))
        if not is_success(self._response.status_code):
            raise GoogleCloudServiceError(self._response.json())

    @property
    def _servie_url(self):
        return f'{self.SERVICE_ENDPOINT}{self.SERVICE_ENDPOINT_URI}'

    def _create_task(self, service, uri):
        self._task = {
            'task': {
                'appEngineHttpRequest': {
                    'appEngineRouting': {
                        'service': service
                    },
                    'relativeUri': uri,
                    'headers': {'Content-Type': 'application/json'},
                },
            },
        }

    def _add_access_token(self, user):
        get_user_model = import_string('django.contrib.auth.get_user_model')
        refresh_token_class = import_string('rest_framework_simplejwt.tokens.RefreshToken')
        user_class = get_user_model()
        user = None if user is None else user if isinstance(user, user_class) else user_class.objects.get(pk=user)
        if user:
            access_token = refresh_token_class.for_user(user).access_token
            headers = self._task['task']['appEngineHttpRequest']['headers']
            headers.update({'X-Forwarded-Authorization': f'Bearer {access_token}'})

    def _add_http_method(self, http_method):
        self._task['task']['appEngineHttpRequest'].update({'httpMethod': http_method})

    def _add_body(self, payload: dict):
        payload_json = json.dumps(payload)
        payload_string_bytes = payload_json.encode('ascii')
        payload_bytes = base64.b64encode(payload_string_bytes)
        body = payload_bytes.decode('ascii')
        self._task['task']['appEngineHttpRequest'].update({'body': body})



"""
py setup.py sdist
twine upload --repository-url https://europe-west1-python.pkg.dev/expressmoney/default/ dist/expressmoney-13.1.1.tar.gz
twine upload dist/expressmoney-13.1.1.tar.gz
"""
import setuptools

setuptools.setup(
    name='expressmoney',
    packages=setuptools.find_packages(),
    version='13.1.1',
    description='SDK ExpressMoney',
    author='Development team',
    author_email='dev@expressmoney.com',
    install_requires=('requests', 'google-auth', ),
    python_requires='>=3.7',
)
